@@ -47,7 +47,6 @@
 
 #define BS_SG_RESVD_SZ  (512 * 1024)
 #define BS_SG_TIMEOUT	2000
-#define BS_SG_SHIFT	9
 
 static int graceful_read(int fd, void *p_read, int to_read)
 {
@@ -270,6 +269,24 @@ static void bs_sg_cmd_complete(int fd, int events, void *data)
 	cmd->scsi_cmd_done(cmd, io_hdr.status);
 }
 
+static int get_bsg_major(char *path)
+{
+	FILE *devfd;
+	char majorno[8];
+	char dev[64];
+	char tmp[16];
+
+	sscanf(path, "/dev/bsg/%s", tmp);
+	sprintf(dev, "/sys/class/bsg/%s/dev", tmp);
+	devfd = fopen(dev, "r");
+	if (!devfd)
+		return -1;
+	fscanf(devfd, "%s:", majorno);
+	fclose(devfd);
+
+	return atoi(majorno);
+}
+
 static int chk_sg_device(char *path)
 {
 	struct stat st;
@@ -288,9 +305,10 @@ static int chk_sg_device(char *path)
 	if (major(st.st_rdev) == SCSI_GENERIC_MAJOR)
 		return 0;
 
-	/* This is not yet defined in include/linux/major.h.. */
-	if (major(st.st_rdev) == 254)
-		return 1;
+	if (!strncmp("/dev/bsg", path, 8)) {
+		if (major(st.st_rdev) == get_bsg_major(path))
+			return 1;
+	}
 
 	return -1;
 }
@@ -373,7 +391,7 @@ static int bs_sg_open(struct scsi_lu *lu, char *path, int *fd, uint64_t *size)
 		cmd_complete = &bs_bsg_cmd_complete;
 		err = init_bsg_device(sg_fd);
 	} else {
-		cmd_complete = bs_sg_cmd_complete;
+		cmd_complete = &bs_sg_cmd_complete;
 		err = init_sg_device(sg_fd);
 	}
 	if (err) {
