@@ -74,13 +74,6 @@ enum {
 	IOSTATE_TX_END,
 };
 
-int iscsi_rdma_enabled;
-
-int iscsi_pthread_per_target(void)
-{
-	return sig_fd >= 0 && !iscsi_rdma_enabled;
-}
-
 void conn_read_pdu(struct iscsi_connection *conn)
 {
 	conn->rx_iostate = IOSTATE_RX_BHS;
@@ -437,6 +430,8 @@ static void login_start(struct iscsi_connection *conn)
 	struct iscsi_login_rsp *rsp = (struct iscsi_login_rsp *)&conn->rsp.bhs;
 	char *name, *alias, *session_type, *target_name;
 	struct iscsi_target *target;
+	char buf[NI_MAXHOST + NI_MAXSERV + 4];
+	int reason, redir;
 
 	conn->cid = be16_to_cpu(req->cid);
 	memcpy(conn->isid, req->isid, sizeof(req->isid));
@@ -493,16 +488,28 @@ static void login_start(struct iscsi_connection *conn)
 			conn->state = STATE_EXIT;
 			return;
 		}
+		if (target->rdma) {
+			eprintf("Target %s is RDMA, but conn cid:%d from %s is TCP\n",
+				target_name, conn->cid, conn->initiator);
+			rsp->status_class = ISCSI_STATUS_CLS_INITIATOR_ERR;
+			rsp->status_detail = ISCSI_LOGIN_STATUS_TGT_NOT_FOUND;
+			conn->state = STATE_EXIT;
+			return;
+		}
 
 		conn->tid = target->tid;
 
-		if (target_redirected(target, conn)) {
-			char buf[NI_MAXHOST + NI_MAXSERV + 4];
-			snprintf(buf, sizeof(buf), "%s:%s", target->redirect_info.addr,
-				 target->redirect_info.port);
+		redir = target_redirected(target, conn, buf, &reason);
+		if (redir < 0) {
+			rsp->status_class = ISCSI_STATUS_CLS_TARGET_ERR;
+			rsp->status_detail = ISCSI_LOGIN_STATUS_TARGET_ERROR;
+			conn->state = STATE_EXIT;
+			return;
+		}
+		else if (redir) {
 			text_key_add(conn, "TargetAddress", buf);
 			rsp->status_class = ISCSI_STATUS_CLS_REDIRECT;
-			rsp->status_detail = target->redirect_info.reason;
+			rsp->status_detail = reason;
 			conn->state = STATE_EXIT;
 			return;
 		}
@@ -1925,8 +1932,21 @@ static int do_recv(struct iscsi_connection *conn, int next_state)
 			return -EIO;
 	}
 
+	conn->stats.rxdata_octets += ret;
 	conn->rx_size -= ret;
 	conn->rx_buffer += ret;
+
+	if (conn->rx_iostate == IOSTATE_RX_BHS) {
+		switch (conn->req.bhs.opcode & ISCSI_OPCODE_MASK) {
+		case ISCSI_OP_SCSI_CMD:
+			conn->stats.scsicmd_pdus++;
+			break;
+		case ISCSI_OP_SCSI_DATA_OUT:
+			conn->stats.dataout_pdus++;
+		}
+	}
+
+
 	if (!conn->rx_size)
 		conn->rx_iostate = next_state;
 
@@ -1938,6 +1958,7 @@ void iscsi_rx_handler(struct iscsi_connection *conn)
 	int ret = 0, hdigest, ddigest;
 	uint32_t crc;
 
+
 	if (conn->state == STATE_SCSI) {
 		struct param *p = conn->session_param;
 		hdigest = p[ISCSI_PARAM_HDRDGST_EN].val & DIGEST_CRC32C;
@@ -2093,8 +2114,21 @@ again:
 		return -EIO;
 	}
 
+	conn->stats.txdata_octets += ret;
 	conn->tx_size -= ret;
 	conn->tx_buffer += ret;
+
+	if (conn->tx_iostate == IOSTATE_TX_BHS) {
+		switch (conn->rsp.bhs.opcode) {
+		case ISCSI_OP_SCSI_DATA_IN:
+			conn->stats.datain_pdus++;
+			break;
+		case ISCSI_OP_SCSI_CMD_RSP:
+			conn->stats.scsirsp_pdus++;
+			break;
+		}
+	}
+
 	if (conn->tx_size)
 		goto again;
 	conn->tx_iostate = next_state;
@@ -2247,7 +2281,7 @@ finish:
 		else {
 			conn->state = STATE_SCSI;
 			conn_read_pdu(conn);
-			conn->tp->ep_nexus_init(conn);
+			conn->tp->ep_event_modify(conn, EPOLLIN);
 		}
 		break;
 	case STATE_EXIT:
@@ -2266,7 +2300,7 @@ out:
 	return ret;
 }
 
-static int iscsi_transportid(int tid, uint64_t itn_id, char *buf, int size)
+int iscsi_transportid(int tid, uint64_t itn_id, char *buf, int size)
 {
 	struct iscsi_session *session;
 	char *p;
@@ -2293,7 +2327,7 @@ static int iscsi_transportid(int tid, uint64_t itn_id, char *buf, int size)
 
 	put_unaligned_be16(len - 4, buf + 2);
 
-	sprintf(buf + 4, session->initiator);
+	sprintf(buf + 4, "%s", session->initiator);
 
 	p = buf + (4 + strlen(session->initiator) + 1);
 
