@@ -197,6 +197,8 @@ struct iscsi_connection {
 	} auth;
 
 	struct iscsi_transport *tp;
+
+	struct iscsi_stats stats;
 };
 
 #define STATE_FREE		0
@@ -212,6 +214,9 @@ struct iscsi_connection {
 #define STATE_CLOSE		10
 #define STATE_EXIT		11
 #define STATE_SCSI		12
+#define STATE_INIT		13
+#define STATE_START		14
+#define STATE_READY		15
 
 #define AUTH_STATE_START	0
 #define AUTH_STATE_CHALLENGE	1
@@ -248,16 +253,12 @@ struct iscsi_target {
 		char addr[NI_MAXHOST + 1];
 		char port[NI_MAXSERV + 1];
 		uint8_t reason;
+		char	*callback;
 	} redirect_info;
 
 	struct list_head isns_list;
 
-	int efd;
-	pthread_mutex_t event_lock;
-	struct list_head events_list;
-
-	struct bs_finish bsfin;
-	int stop_pthread;
+	int rdma;
 };
 
 enum task_flags {
@@ -300,6 +301,7 @@ extern void conn_read_pdu(struct iscsi_connection *conn);
 extern int iscsi_tx_handler(struct iscsi_connection *conn);
 extern void iscsi_rx_handler(struct iscsi_connection *conn);
 extern int iscsi_scsi_cmd_execute(struct iscsi_task *task);
+extern int iscsi_transportid(int tid, uint64_t itn_id, char *buf, int size);
 
 /* iscsid.c iscsi_task */
 extern void iscsi_free_task(struct iscsi_task *task);
@@ -323,9 +325,8 @@ extern int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid,
 			     uint64_t lun, char *buf, int rest);
 int iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
 			uint32_t cid, char *name);
-int target_redirected(struct iscsi_target *target, struct iscsi_connection *conn);
-
-int iscsi_pthread_per_target(void);
+int target_redirected(struct iscsi_target *target,
+	struct iscsi_connection *conn, char *buf, int *reason);
 
 /* param.c */
 int param_index_by_name(char *name, struct iscsi_key *keys);
