@@ -43,6 +43,8 @@
 #define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 #define BUFSIZE (1 << 18)
 
+#define EID_NAME_KEY "eid"
+
 struct isns_io {
 	char *buf;
 	int offset;
@@ -67,13 +69,36 @@ static int use_isns, use_isns_ac, isns_fd, scn_listen_fd, scn_fd;
 static struct isns_io isns_rx, scn_rx;
 static char *rxbuf;
 static uint16_t transaction;
-static uint32_t isns_timeout = 30; /* seconds */
 static char eid[ISCSI_NAME_LEN];
 static uint8_t ip[16]; /* IET supoprts only one portal */
 static struct sockaddr_storage ss;
 
+/*
+ * Section 6.2.6
+ * The registration SHALL be removed from the iSNS database if an iSNS
+ * Protocol message is not received from the iSNS client before the
+ * registration period has expired. Receipt of any iSNS Protocol
+ * message from the iSNS client automatically refreshes the Entity
+ * Registration Period and Entity Registration Timestamp. To prevent a
+ * registration from expiring, the iSNS client should send an iSNS
+ * Protocol message to the iSNS server at intervals shorter than the
+ * registration period.
+ *
+ * Implementor's Note:
+ * We send a DevAttrQry message to the iSNS server every isns_timeout
+ * seconds to keep our entries alive at the iSNS server.
+ * To start with, we assume that the registration period is greater
+ * than 30 seconds. After we register an entity we query for the
+ * registration period and change the isns_timeout to be 10 seconds less
+ * than the regiistration period. If the iSNS server doesn't respond
+ * with a registration period we stick with 30 seconds.
+ */
+#define DEFAULT_ISNS_TIMEOUT 30 /* seconds */
+static uint32_t isns_timeout = DEFAULT_ISNS_TIMEOUT;
+
 static char isns_addr[NI_MAXHOST];
 static int isns_port = ISNS_PORT;
+static int num_targets = 0;
 
 int isns_scn_access(int tid, char *name)
 {
@@ -107,9 +132,9 @@ static int isns_get_ip(int fd)
 	}
 
 	err = getnameinfo((struct sockaddr *) &lss, sizeof(lss),
-			  eid, sizeof(eid), NULL, 0, 0);
+			  eid, sizeof(eid), NULL, 0, NI_NUMERICHOST);
 	if (err) {
-		eprintf("getaddrinfo error %s!\n", gai_strerror(err));
+		eprintf("getnameinfo error %s!\n", gai_strerror(err));
 		return err;
 	}
 
@@ -243,15 +268,14 @@ static int isns_scn_deregister(char *name)
 #define set_scn_flag(x)
 #endif
 
-static int isns_scn_register(void)
+static int isns_scn_register(char *name)
 {
 	int err;
 	uint16_t flags, length = 0;
 	uint32_t scn_flags;
-	char buf[4096], *name;
+	char buf[4096];
 	struct isns_hdr *hdr = (struct isns_hdr *) buf;
 	struct isns_tlv *tlv;
-	struct iscsi_target *target;
 
 	if (list_empty(&iscsi_targets_list))
 		return 0;
@@ -263,10 +287,6 @@ static int isns_scn_register(void)
 	memset(buf, 0, sizeof(buf));
 	tlv = (struct isns_tlv *) hdr->pdu;
 
-	target = list_first_entry(&iscsi_targets_list,
-				  struct iscsi_target, tlist);
-	name = tgt_targetname(target->tid);
-
 	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME, name);
 	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME, name);
 	length += isns_tlv_set(&tlv, 0, 0, 0);
@@ -289,6 +309,52 @@ static int isns_scn_register(void)
 	return 0;
 }
 
+static int isns_eid_attr_query(void)
+{
+	int err;
+	uint16_t flags, length = 0;
+	char buf[4096];
+	struct isns_hdr *hdr = (struct isns_hdr *) buf;
+	struct isns_tlv *tlv;
+	struct iscsi_target *target;
+	struct isns_qry_mgmt *mgmt;
+	char *name;
+
+	eprintf("\n");
+	if (!isns_fd)
+		if (isns_connect() < 0)
+			return 0;
+
+	mgmt = malloc(sizeof(*mgmt));
+	if (!mgmt)
+		return 0;
+	list_add(&mgmt->qlist, &qry_list);
+
+	memset(buf, 0, sizeof(buf));
+	tlv = (struct isns_tlv *) hdr->pdu;
+
+	strcpy(mgmt->name, EID_NAME_KEY);
+	target = list_first_entry(&iscsi_targets_list,
+				  struct iscsi_target, tlist);
+	name = tgt_targetname(target->tid);
+
+	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME, name);
+	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ENTITY_IDENTIFIER, eid);
+	length += isns_tlv_set(&tlv, 0, 0, 0);
+	length += isns_tlv_set(&tlv, ISNS_ATTR_REGISTRATION_PERIOD, 0, 0);
+
+	flags = ISNS_FLAG_CLIENT | ISNS_FLAG_LAST_PDU | ISNS_FLAG_FIRST_PDU;
+	isns_hdr_init(hdr, ISNS_FUNC_DEV_ATTR_QRY, length, flags,
+		      ++transaction, 0);
+	mgmt->transaction = transaction;
+
+	err = write(isns_fd, buf, length + sizeof(struct isns_hdr));
+	if (err < 0)
+		eprintf("%d %m\n", length);
+
+	return 0;
+}
+
 static int isns_attr_query(char *name)
 {
 	int err;
@@ -344,59 +410,17 @@ static int isns_attr_query(char *name)
 	return 0;
 }
 
-static int isns_deregister(void)
-{
-	int err;
-	uint16_t flags, length = 0;
-	char buf[4096], *name;
-	struct isns_hdr *hdr = (struct isns_hdr *) buf;
-	struct isns_tlv *tlv;
-	struct iscsi_target *target;
-
-	if (list_empty(&iscsi_targets_list))
-		return 0;
-
-	if (!isns_fd)
-		if (isns_connect() < 0)
-			return 0;
-
-	memset(buf, 0, sizeof(buf));
-	tlv = (struct isns_tlv *) hdr->pdu;
-
-	target = list_first_entry(&iscsi_targets_list,
-				  struct iscsi_target, tlist);
-	name = tgt_targetname(target->tid);
-
-	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME, name);
-	length += isns_tlv_set(&tlv, 0, 0, 0);
-	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ENTITY_IDENTIFIER, eid);
-
-	flags = ISNS_FLAG_CLIENT | ISNS_FLAG_LAST_PDU | ISNS_FLAG_FIRST_PDU;
-	isns_hdr_init(hdr, ISNS_FUNC_DEV_DEREG, length, flags,
-		      ++transaction, 0);
-
-	err = write(isns_fd, buf, length + sizeof(struct isns_hdr));
-	if (err < 0)
-		eprintf("%d %m\n", length);
-	return 0;
-}
-
-static inline int list_length_is_one(const struct list_head *head)
-{
-        return (!list_empty(head) && head->next == head->prev);
-}
-
 int isns_target_register(char *name)
 {
 	char buf[4096];
 	uint16_t flags = 0, length = 0;
 	struct isns_hdr *hdr = (struct isns_hdr *) buf;
 	struct isns_tlv *tlv;
+	struct iscsi_target *target;
 	uint32_t port = htonl(iscsi_listen_port);
 	uint32_t node = htonl(ISNS_NODE_TARGET);
 	uint32_t type = htonl(2);
-	struct iscsi_target *target;
-	int err, initial = list_length_is_one(&iscsi_targets_list);
+	int err;
 
 	if (!use_isns)
 		return 0;
@@ -408,16 +432,16 @@ int isns_target_register(char *name)
 	memset(buf, 0, sizeof(buf));
 	tlv = (struct isns_tlv *) hdr->pdu;
 
-        target = list_first_entry(&iscsi_targets_list,
-				  struct iscsi_target, tlist);
-        length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME,
-				      tgt_targetname(target->tid));
+	target = list_first_entry(&iscsi_targets_list,
+				struct iscsi_target, tlist);
+	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME,
+					tgt_targetname(target->tid));
 	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ENTITY_IDENTIFIER, eid);
 
 	length += isns_tlv_set(&tlv, 0, 0, 0);
 	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ENTITY_IDENTIFIER, eid);
 
-	if (initial) {
+	if (!num_targets) {
 		length += isns_tlv_set(&tlv, ISNS_ATTR_ENTITY_PROTOCOL,
 				       sizeof(type), &type);
 		length += isns_tlv_set(&tlv, ISNS_ATTR_PORTAL_IP_ADDRESS,
@@ -431,6 +455,7 @@ int isns_target_register(char *name)
 			length += isns_tlv_set(&tlv, ISNS_ATTR_SCN_PORT,
 					       sizeof(sport), &sport);
 		}
+		add_work(&timeout_work, DEFAULT_ISNS_TIMEOUT);
 	}
 
 	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME, name);
@@ -446,9 +471,12 @@ int isns_target_register(char *name)
 		eprintf("%d %m\n", length);
 
 	if (scn_listen_port)
-		isns_scn_register();
+		isns_scn_register(name);
 
+	if (!num_targets)
+		isns_eid_attr_query();
 	isns_attr_query(name);
+	num_targets++;
 
 	return 0;
 }
@@ -460,6 +488,7 @@ static void free_all_acl(struct iscsi_target *target)
 	while (!list_empty(&target->isns_list)) {
 		ini = list_first_entry(&target->isns_list, typeof(*ini), ilist);
 		list_del(&ini->ilist);
+		free(ini);
 	}
 }
 
@@ -469,7 +498,7 @@ int isns_target_deregister(char *name)
 	uint16_t flags, length = 0;
 	struct isns_hdr *hdr = (struct isns_hdr *) buf;
 	struct isns_tlv *tlv;
-	int err, last = list_empty(&iscsi_targets_list);
+	int err;
 	struct iscsi_target *target;
 
 	target = target_find_by_name(name);
@@ -483,6 +512,7 @@ int isns_target_deregister(char *name)
 		if (isns_connect() < 0)
 			return 0;
 
+	num_targets--;
 	isns_scn_deregister(name);
 
 	memset(buf, 0, sizeof(buf));
@@ -490,10 +520,11 @@ int isns_target_deregister(char *name)
 
 	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME, name);
 	length += isns_tlv_set(&tlv, 0, 0, 0);
-	if (last)
+	if (!num_targets) {
+		del_work(&timeout_work);
 		length += isns_tlv_set_string(&tlv, ISNS_ATTR_ENTITY_IDENTIFIER,
 					      eid);
-	else
+	} else
 		length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME, name);
 
 	flags = ISNS_FLAG_CLIENT | ISNS_FLAG_LAST_PDU | ISNS_FLAG_FIRST_PDU;
@@ -618,7 +649,7 @@ static char *print_scn_pdu(struct isns_hdr *hdr)
 		case ISNS_ATTR_ISCSI_NAME:
 			eprintf("scn name: %u, %s\n", vlen, (char *) tlv->value);
 			if (!name) {
-				snprintf(iscsi_name, sizeof(iscsi_name), (char *)tlv->value);
+				snprintf(iscsi_name, sizeof(iscsi_name), "%s", (char *)tlv->value);
 				name = iscsi_name;
 			}
 			break;
@@ -644,9 +675,10 @@ static void qry_rsp_handle(struct isns_hdr *hdr)
 	uint16_t function, length, flags, transaction, sequence;
 	uint32_t status = (uint32_t) (*hdr->pdu);
 	struct isns_qry_mgmt *mgmt, *n;
-	struct iscsi_target *target;
+	struct iscsi_target *target = NULL;
 	struct isns_initiator *ini;
 	char *name = NULL;
+	int reg_period = 0;
 
 	get_hdr_param(hdr, function, length, flags, transaction, sequence);
 
@@ -671,13 +703,15 @@ found:
 		goto free_qry_mgmt;
 	}
 
-	target = target_find_by_name(mgmt->name);
-	if (!target) {
-		eprintf("invalid tid %s\n", mgmt->name);
-		goto free_qry_mgmt;
-	}
+	if (strcmp(mgmt->name, EID_NAME_KEY)) {
+		target = target_find_by_name(mgmt->name);
+		if (!target) {
+			eprintf("invalid tid %s\n", mgmt->name);
+			goto free_qry_mgmt;
+		}
 
-	free_all_acl(target);
+		free_all_acl(target);
+	}
 
 	/* skip status */
 	tlv = (struct isns_tlv *) ((char *) hdr->pdu + 4);
@@ -707,6 +741,8 @@ found:
 			} else
 				name = NULL;
 			break;
+		case ISNS_ATTR_REGISTRATION_PERIOD:
+			reg_period = ntohl(*(tlv->value));
 		default:
 			name = NULL;
 			break;
@@ -716,6 +752,14 @@ found:
 		tlv = (struct isns_tlv *) ((char *) tlv->value + vlen);
 	}
 
+	/*
+	 * If the iSNS server provided a registration period, change
+	 * isns_timeout to be slighly less than the period, to make
+	 * sure our registrations are not kicked out
+	 */
+	if (reg_period)
+		isns_timeout = reg_period - 10;
+
 free_qry_mgmt:
 	free(mgmt);
 }
@@ -751,6 +795,7 @@ static void isns_handle(int fd, int events, void *data)
 		break;
 	case ISNS_FUNC_DEV_DEREG_RSP:
 	case ISNS_FUNC_SCN_REG_RSP:
+	case ISNS_FUNC_SCN_DEREG_RSP:
 		break;
 	case ISNS_FUNC_SCN:
 		name = print_scn_pdu(hdr);
@@ -915,6 +960,7 @@ int isns_init(void)
 	int err;
 	char port[8];
 	struct addrinfo hints, *res;
+	struct iscsi_target *target;
 
 	snprintf(port, sizeof(port), "%d", isns_port);
 	memset(&hints, 0, sizeof(hints));
@@ -924,7 +970,7 @@ int isns_init(void)
 		eprintf("getaddrinfo error %s\n", isns_addr);
 		return -1;
 	}
-	memcpy(&ss, res->ai_addr, sizeof(ss));
+	memcpy(&ss, res->ai_addr, sizeof(*res->ai_addr));
 	freeaddrinfo(res);
 
 	rxbuf = calloc(2, BUFSIZE);
@@ -942,13 +988,49 @@ int isns_init(void)
 
 	use_isns = 1;
 
+	if (!num_targets)
+		list_for_each_entry(target, &iscsi_targets_list, tlist)
+			isns_target_register(tgt_targetname(target->tid));
+
 	timeout_work.func = isns_timeout_fn;
 	timeout_work.data = &timeout_work;
-	add_work(&timeout_work, isns_timeout);
 
 	return 0;
 }
 
+int isns_eid_deregister(void)
+{
+	char buf[4096];
+	uint16_t flags, length = 0;
+	struct isns_hdr *hdr = (struct isns_hdr *) buf;
+	struct isns_tlv *tlv;
+	struct iscsi_target *target;
+	int err;
+
+	if (!isns_fd)
+		if (isns_connect() < 0)
+			return 0;
+
+	memset(buf, 0, sizeof(buf));
+	tlv = (struct isns_tlv *) hdr->pdu;
+
+	target = list_first_entry(&iscsi_targets_list,
+				struct iscsi_target, tlist);
+	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ISCSI_NAME,
+					tgt_targetname(target->tid));
+	length += isns_tlv_set(&tlv, 0, 0, 0);
+	length += isns_tlv_set_string(&tlv, ISNS_ATTR_ENTITY_IDENTIFIER,
+					      eid);
+	flags = ISNS_FLAG_CLIENT | ISNS_FLAG_LAST_PDU | ISNS_FLAG_FIRST_PDU;
+	isns_hdr_init(hdr, ISNS_FUNC_DEV_DEREG, length, flags,
+		      ++transaction, 0);
+
+	err = write(isns_fd, buf, length + sizeof(struct isns_hdr));
+	if (err < 0)
+		eprintf("%d %m\n", length);
+
+	return 0;
+}
 void isns_exit(void)
 {
 	struct iscsi_target *target;
@@ -956,10 +1038,13 @@ void isns_exit(void)
 	if (!use_isns)
 		return;
 
-	list_for_each_entry(target, &iscsi_targets_list, tlist)
-		isns_scn_deregister(tgt_targetname(target->tid));
+	if (num_targets) {
+		del_work(&timeout_work);
+		list_for_each_entry(target, &iscsi_targets_list, tlist)
+			free_all_acl(target);
 
-	isns_deregister();
+		isns_eid_deregister();
+	}
 
 	if (isns_fd) {
 		tgt_event_del(isns_fd);
@@ -974,6 +1059,7 @@ void isns_exit(void)
 		close(scn_fd);
 	}
 
+	num_targets = use_isns = isns_fd = scn_listen_fd = scn_fd = 0;
 	free(rxbuf);
 }
 
@@ -1013,7 +1099,7 @@ int isns_update(char *params)
 
 	while ((p = strsep(&params, ",")) != NULL) {
 		substring_t args[MAX_OPT_ARGS];
-		int token;
+		int token, isns;
 		char tmp[16];
 
 		if (!*p)
@@ -1022,16 +1108,21 @@ int isns_update(char *params)
 
 		switch (token) {
 		case Opt_isns:
-			/* FIXME */
-			if (use_isns) {
+			match_strncpy(tmp, &args[0], sizeof(tmp));
+			if (!strcmp(tmp, "On"))
+				isns = 1;
+			else if (!strcmp(tmp, "Off"))
+				isns = 0;
+			else {
 				ret = TGTADM_INVALID_REQUEST;
 				break;
 			}
-
-			match_strncpy(tmp, &args[0], sizeof(tmp));
-			use_isns = !strcmp(tmp, "On");
-			if (use_isns)
+			if (use_isns == isns)
+				break;
+			if (isns)
 				isns_init();
+			else
+				isns_exit();
 			break;
 		case Opt_ip:
 			match_strncpy(isns_addr, &args[0], sizeof(isns_addr));
