@@ -25,7 +25,6 @@
 #include <unistd.h>
 #include <netdb.h>
 #include <sys/stat.h>
-#include <sys/epoll.h>
 #include <sys/un.h>
 #include <netinet/in.h>
 #include <sys/socket.h>
@@ -33,12 +32,10 @@
 #include <netinet/tcp.h>
 #include <netinet/ip.h>
 #include <arpa/inet.h>
-#include <pthread.h>
 #include "iscsid.h"
 #include "tgtadm.h"
 #include "tgtd.h"
 #include "target.h"
-#include "util.h"
 
 LIST_HEAD(iscsi_targets_list);
 
@@ -191,28 +188,98 @@ int ip_acl(int tid, struct iscsi_connection *conn)
 	}
 	return -EPERM;
 }
+static int
+get_redirect_address(char *callback, char *buffer, int buflen,
+			char **address, char **ip_port, int *rsn)
+{
+	char *p, *addr, *port;
 
-int target_redirected(struct iscsi_target *target, struct iscsi_connection *conn)
+	bzero(buffer, buflen);
+	if (call_program(callback, NULL, NULL, buffer, buflen, 0))
+		return -1;
+
+	/* syntax is string_addr:string_port:string_reason */
+	addr = p = buffer;
+	if (*p == '[') {
+		while (*p != ']' && *p != '\0')
+			p++;
+		if (*p == ']') {
+			p++;
+			if (*p != ':')
+				return -1;
+		}
+	} else {
+		while (*p != ':' && *p != '\0')
+			p++;
+	}
+	if (!*p)
+		return -1;
+	*p = '\0';
+	port = ++p;
+	while (*p != ':' && *p != '\0')
+		p++;
+	if (!*p)
+		return -1;
+	*p = '\0';
+	p++;
+	if (!strncmp(p, "Temporary", 9))
+		*rsn = ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP;
+	else if (!strncmp(p, "Permanent", 9))
+		*rsn = ISCSI_LOGIN_STATUS_TGT_MOVED_PERM;
+	else
+		return -1;
+	*address = addr;
+	*ip_port = port;
+	return 0;
+}
+
+int target_redirected(struct iscsi_target *target,
+	struct iscsi_connection *conn, char *buf, int *reason)
 {
 	struct sockaddr_storage from;
 	struct addrinfo hints, *res;
 	socklen_t len;
-	int ret;
-	char *p, *q, *str;
-
-	if (!strlen(target->redirect_info.addr))
-		return 0;
-
-	if (target->redirect_info.reason != ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP &&
-	    target->redirect_info.reason != ISCSI_LOGIN_STATUS_TGT_MOVED_PERM)
-		return 0;
+	int ret, rsn = 0;
+	char *p, *q, *str, *port = NULL, *addr;
+	char buffer[NI_MAXHOST + NI_MAXSERV + 4];
+	char dst[INET6_ADDRSTRLEN], in_buf[1024];
 
 	len = sizeof(from);
 	ret = conn->tp->ep_getpeername(conn, (struct sockaddr *)&from, &len);
 	if (ret < 0)
 		return 0;
 
-	p = strdup(target->redirect_info.addr);
+	ret = 1;
+	if (target->redirect_info.callback) {
+		p = in_buf;
+		p += sprintf(p, "%s ", target->redirect_info.callback);
+		p += sprintf(p, "%s ", tgt_targetname(target->tid));
+		ret = getnameinfo((struct sockaddr *)&from, sizeof(from), dst,
+				sizeof(dst), NULL, 0, NI_NUMERICHOST);
+		if (ret)
+			goto predefined;
+		sprintf(p, "%s", dst);
+		ret = get_redirect_address(in_buf, buffer,
+					sizeof(buffer), &addr, &port, &rsn);
+		if (ret)
+			return -1;
+	}
+
+predefined:
+	if (ret) {
+		if (!strlen(target->redirect_info.addr))
+			return 0;
+
+		addr = target->redirect_info.addr;
+		port = target->redirect_info.port;
+		rsn = target->redirect_info.reason;
+	}
+
+	if (rsn != ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP &&
+	    rsn != ISCSI_LOGIN_STATUS_TGT_MOVED_PERM)
+		return 0;
+
+	p = strdup(addr);
 	if (!p)
 		return 0;
 	str = p;
@@ -240,6 +307,10 @@ int target_redirected(struct iscsi_target *target, struct iscsi_connection *conn
 	freeaddrinfo(res);
 	free(str);
 
+	if (!ret) {
+		sprintf(buf, "%s:%s", addr, port);
+		*reason = rsn;
+	}
 	return !ret;
 }
 
@@ -306,63 +377,14 @@ void iscsi_target_destroy(int tid)
 	}
 
 	list_del(&target->tlist);
-
-	pthread_mutex_init(&target->event_lock, NULL);
-
-	if (target->bsfin.thread) {
-		target->stop_pthread = 1;
-		pthread_kill(target->bsfin.thread, SIGUSR2);
-
-		pthread_join(target->bsfin.thread, NULL);
-		pthread_mutex_destroy(&target->bsfin.finished_lock);
-	}
-
-	close(target->efd);
+	if (target->redirect_info.callback)
+		free(target->redirect_info.callback);
 	free(target);
 	isns_target_deregister(tgt_targetname(tid));
 
 	return;
 }
 
-static void *iscsi_thread_fn(void *arg)
-{
-	struct iscsi_target *t = arg;
-	struct epoll_event events[1024];
-	struct event_data *tev;
-	sigset_t mask;
-	int nevent, i;
-
-	sigemptyset(&mask);
-	sigaddset(&mask, SIGUSR2);
-	pthread_sigmask(SIG_BLOCK, &mask, NULL);
-
-	pthread_mutex_lock(&t->event_lock);
-
-	do_tgt_event_add(t->efd, &t->events_list, sig_fd, EPOLLIN,
-			 bs_sig_request_done, &t->bsfin);
-
-	pthread_mutex_unlock(&t->event_lock);
-
-retry:
-	nevent = epoll_wait(t->efd, events, ARRAY_SIZE(events), 1000);
-	if (nevent < 0) {
-		if (errno != EINTR) {
-			eprintf("%m\n");
-			exit(1);
-		}
-	} else if (nevent) {
-		for (i = 0; i < nevent; i++) {
-			tev = (struct event_data *) events[i].data.ptr;
-			tev->handler(tev->fd, events[i].events, tev->data);
-		}
-	}
-
-	if (!t->stop_pthread)
-		goto retry;
-
-	pthread_exit(NULL);
-}
-
 int iscsi_target_create(struct target *t)
 {
 	int tid = t->tid;
@@ -393,15 +415,11 @@ int iscsi_target_create(struct target *t)
 		[ISCSI_PARAM_MAX_OUTST_PDU] =  {0, 0},  /* not in open-iscsi */
 	};
 
-	target = zalloc(sizeof(*target));
+	target = malloc(sizeof(*target));
 	if (!target)
 		return -ENOMEM;
 
-	target->efd = epoll_create(128);
-	if (target->efd < 0) {
-		free(target);
-		return -EINVAL;
-	}
+	memset(target, 0, sizeof(*target));
 
 	memcpy(target->session_param, default_tgt_session_param,
 	       sizeof(target->session_param));
@@ -409,21 +427,10 @@ int iscsi_target_create(struct target *t)
 	INIT_LIST_HEAD(&target->tlist);
 	INIT_LIST_HEAD(&target->sessions_list);
 	INIT_LIST_HEAD(&target->isns_list);
-	INIT_LIST_HEAD(&target->events_list);
 	target->tid = tid;
 	list_add_tail(&target->tlist, &iscsi_targets_list);
 
 	isns_target_register(tgt_targetname(tid));
-
-	if (iscsi_pthread_per_target()) {
-		pthread_create(&target->bsfin.thread, NULL, iscsi_thread_fn, target);
-
-		pthread_mutex_init(&target->bsfin.finished_lock, NULL);
-		INIT_LIST_HEAD(&target->bsfin.finished_list);
-		t->bsf = &target->bsfin;
-		eprintf("create thread %u\n", (unsigned)target->bsfin.thread);
-	}
-
 	return 0;
 }
 
@@ -488,6 +495,13 @@ int iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
 				err = TGTADM_SUCCESS;
 			} else
 				break;
+		} else if (!strncmp(name, "RedirectCallback", 16)) {
+			target->redirect_info.callback = strdup(str);
+			if (!target->redirect_info.callback) {
+				err = TGTADM_NOMEM;
+				break;
+			}
+			err = TGTADM_SUCCESS;
 		}
 
 		idx = param_index_by_name(name, session_keys);
@@ -522,21 +536,6 @@ static int show_iscsi_param(char *buf, struct param *param, int rest)
 	return total;
 }
 
-static int iscsi_target_show_session(struct iscsi_target* target, uint64_t sid,
-				     char *buf, int rest)
-{
-	int len = 0, total = 0;
-	struct iscsi_session *session;
-
-	list_for_each_entry(session, &target->sessions_list, slist) {
-		if (session->tsih == sid)
-			len = show_iscsi_param(buf, session->session_param, rest);
-			buffer_check(buf, total, len, rest);
-	}
-
-	return total;
-}
-
 #define __buffer_check(buf, total, len, rest)	\
 ({						\
 	buf += len;				\
@@ -546,6 +545,70 @@ static int iscsi_target_show_session(struct iscsi_target* target, uint64_t sid,
 		return total;			\
 })
 
+static struct iscsi_session *iscsi_target_find_session(
+	struct iscsi_target *target,
+	uint64_t sid)
+{
+	struct iscsi_session *session;
+
+	list_for_each_entry(session, &target->sessions_list, slist) {
+		if (session->tsih == sid)
+			return session;
+	}
+
+	return NULL;
+
+}
+
+static int iscsi_target_show_session(struct iscsi_target *target, uint64_t sid,
+				     char *buf, int rest)
+{
+	int len = 0, total = 0;
+	struct iscsi_session *session;
+
+	session = iscsi_target_find_session(target, sid);
+
+	if (session) {
+		len = show_iscsi_param(buf, session->session_param, rest);
+		__buffer_check(buf, total, len, rest);
+
+	}
+
+	return total;
+}
+
+static int iscsi_target_show_stats(struct iscsi_target *target, uint64_t sid,
+				   char *buf, int rest)
+{
+	int len = 0, total = 0;
+	struct iscsi_session *session;
+	struct iscsi_connection *conn;
+
+	session = iscsi_target_find_session(target, sid);
+
+	if (session) {
+		list_for_each_entry(conn, &session->conn_list, clist) {
+			len = snprintf(buf, rest,
+				       "rxdata_octets: %" PRIu64 "\n"
+				       "txdata_octets: %" PRIu64 "\n"
+				       "dataout_pdus:  %d\n"
+				       "datain_pdus:   %d\n"
+				       "scsicmd_pdus:  %d\n"
+				       "scsirsp_pdus:  %d\n",
+				       conn->stats.rxdata_octets,
+				       conn->stats.txdata_octets,
+				       conn->stats.dataout_pdus,
+				       conn->stats.datain_pdus,
+				       conn->stats.scsicmd_pdus,
+				       conn->stats.scsirsp_pdus);
+			__buffer_check(buf, total, len, rest);
+		}
+	}
+
+	return total;
+
+}
+
 static int show_redirect_info(struct iscsi_target* target, char *buf, int rest)
 {
 	int len, total = 0;
@@ -568,6 +631,16 @@ static int show_redirect_info(struct iscsi_target* target, char *buf, int rest)
 	return total;
 }
 
+static int show_redirect_callback(struct iscsi_target *target, char *buf, int rest)
+{
+	int len, total = 0;
+
+	len = snprintf(buf, rest, "RedirectCallback=%s\n", target->redirect_info.callback);
+	__buffer_check(buf, total, len, rest);
+
+	return total;
+}
+
 int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lun,
 		      char *buf, int rest)
 {
@@ -585,7 +658,9 @@ int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lu
 		total = isns_show(buf, rest);
 		break;
 	case MODE_TARGET:
-		if (strlen(target->redirect_info.addr))
+		if (target->redirect_info.callback)
+			len = show_redirect_callback(target, buf, rest);
+		else if (strlen(target->redirect_info.addr))
 			len = show_redirect_info(target, buf, rest);
 		else
 			len = show_iscsi_param(buf, target->session_param, rest);
@@ -595,6 +670,10 @@ int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lu
 		len = iscsi_target_show_session(target, sid, buf, rest);
 		total += len;
 		break;
+	case MODE_STATS:
+		len = iscsi_target_show_stats(target, sid, buf, rest);
+		total += len;
+		break;
 	default:
 		break;
 	}
