@@ -41,7 +41,9 @@
 #include "spc.h"
 #include "tgtadm_error.h"
 
-#define BLK_SHIFT	9
+#define DEFAULT_BLK_SHIFT 9
+
+static unsigned int blk_shift = DEFAULT_BLK_SHIFT;
 
 static int sbc_mode_page_update(struct scsi_cmd *cmd, uint8_t *data, int *changed)
 {
@@ -72,19 +74,58 @@ static int sbc_mode_select(int host_no, struct scsi_cmd *cmd)
 	return spc_mode_select(host_no, cmd, sbc_mode_page_update);
 }
 
+static int sbc_mode_sense(int host_no, struct scsi_cmd *cmd)
+{
+	int ret;
+
+	ret = spc_mode_sense(host_no, cmd);
+
+	/*
+	 * If this is a read-only lun, we must modify the data and set the
+	 * write protect bit
+	 */
+	if (cmd->dev->attrs.readonly && ret == SAM_STAT_GOOD) {
+		uint8_t *data, mode6;
+
+		mode6 = (cmd->scb[0] == 0x1a);
+		data = scsi_get_in_buffer(cmd);
+
+		if (mode6)
+			data[2] |= 0x80;
+		else
+			data[3] |= 0x80;
+	}
+
+	return ret;
+}
+
 static int sbc_rw(int host_no, struct scsi_cmd *cmd)
 {
 	int ret;
 	unsigned char key = ILLEGAL_REQUEST;
 	uint16_t asc = ASC_LUN_NOT_SUPPORTED;
+	struct scsi_lu *lu = cmd->dev;
 
 	ret = device_reserved(cmd);
 	if (ret)
 		return SAM_STAT_RESERVATION_CONFLICT;
 
+	if (lu->attrs.readonly) {
+		switch (cmd->scb[0]) {
+		case WRITE_6:
+		case WRITE_10:
+		case WRITE_12:
+		case WRITE_16:
+			key = DATA_PROTECT;
+			asc = ASC_WRITE_PROTECT;
+			goto sense;
+			break;
+		}
+	}
+
 	cmd->scsi_cmd_done = target_cmd_io_done;
 
-	cmd->offset = (scsi_rw_offset(cmd->scb) << BLK_SHIFT);
+	cmd->offset = (scsi_rw_offset(cmd->scb) << cmd->dev->blk_shift);
 	ret = cmd->dev->bst->bs_cmd_submit(cmd);
 	if (ret) {
 		key = HARDWARE_ERROR;
@@ -94,6 +135,7 @@ static int sbc_rw(int host_no, struct scsi_cmd *cmd)
 		return SAM_STAT_GOOD;
 	}
 
+sense:
 	cmd->offset = 0;
 	scsi_set_in_resid_by_actual(cmd, 0);
 	scsi_set_out_resid_by_actual(cmd, 0);
@@ -123,6 +165,7 @@ static int sbc_release(int host_no, struct scsi_cmd *cmd)
 static int sbc_read_capacity(int host_no, struct scsi_cmd *cmd)
 {
 	uint32_t *data;
+	unsigned int bshift;
 	uint64_t size;
 	uint8_t *scb = cmd->scb;
 	unsigned char key = ILLEGAL_REQUEST;
@@ -137,11 +180,12 @@ static int sbc_read_capacity(int host_no, struct scsi_cmd *cmd)
 		goto overflow;
 
 	data = scsi_get_in_buffer(cmd);
-	size = cmd->dev->size >> BLK_SHIFT;
+	bshift = cmd->dev->blk_shift;
+	size = cmd->dev->size >> bshift;
 
 	data[0] = (size >> 32) ?
 		__cpu_to_be32(0xffffffff) : __cpu_to_be32(size - 1);
-	data[1] = __cpu_to_be32(1U << BLK_SHIFT);
+	data[1] = __cpu_to_be32(1U << bshift);
 
 overflow:
 	scsi_set_in_resid_by_actual(cmd, 8);
@@ -160,7 +204,9 @@ static int sbc_verify(int host_no, struct scsi_cmd *cmd)
 static int sbc_service_action(int host_no, struct scsi_cmd *cmd)
 {
 	uint32_t *data;
+	unsigned int bshift;
 	uint64_t size;
+	int len = 32;
 
 	if (cmd->scb[1] != SAI_READ_CAPACITY_16)
 		goto sense;
@@ -168,16 +214,19 @@ static int sbc_service_action(int host_no, struct scsi_cmd *cmd)
 	if (scsi_get_in_length(cmd) < 12)
 		goto overflow;
 
-	data = scsi_get_in_buffer(cmd);
-	memset(data, 0, 12);
+	len = min_t(int, len, scsi_get_in_length(cmd));
 
-	size = cmd->dev->size >> BLK_SHIFT;
+	data = scsi_get_in_buffer(cmd);
+	memset(data, 0, len);
+
+	bshift = cmd->dev->blk_shift;
+	size = cmd->dev->size >> bshift;
 
 	*((uint64_t *)(data)) = __cpu_to_be64(size - 1);
-	data[2] = __cpu_to_be32(1UL << BLK_SHIFT);
+	data[2] = __cpu_to_be32(1UL << bshift);
 
 overflow:
-	scsi_set_in_resid_by_actual(cmd, 12);
+	scsi_set_in_resid_by_actual(cmd, len);
 	return SAM_STAT_GOOD;
 sense:
 	sense_data_build(cmd, ILLEGAL_REQUEST, ASC_INVALID_OP_CODE);
@@ -233,11 +282,14 @@ static int sbc_lu_init(struct scsi_lu *lu)
 	lu->attrs.version_desc[2] = 0x0300; /* SPC-3 */
 
 	data = lu->mode_block_descriptor;
-	size = lu->size >> BLK_SHIFT;
+
+	if (!lu->blk_shift)
+		lu->blk_shift = blk_shift; /* if unset, use default shift */
+	size = lu->size >> lu->blk_shift; /* calculate size in blocks */
 
 	*(uint32_t *)(data) = (size >> 32) ?
 			__cpu_to_be32(0xffffffff) : __cpu_to_be32(size);
-	*(uint32_t *)(data + 4) = __cpu_to_be32(1 << BLK_SHIFT);
+	*(uint32_t *)(data + 4) = __cpu_to_be32(1 << lu->blk_shift);
 
 	/* Vendor uniq - However most apps seem to call for mode page 0*/
 	add_mode_page(lu, "0:0:0");
@@ -300,7 +352,7 @@ static struct device_type_template sbc_template = {
 
 		{spc_illegal_op,},
 		{spc_illegal_op,},
-		{spc_mode_sense, NULL, PR_WE_FA|PR_EA_FA|PR_WE_FN|PR_EA_FN},
+		{sbc_mode_sense, NULL, PR_WE_FA|PR_EA_FA|PR_WE_FN|PR_EA_FN},
 		{spc_start_stop, NULL, PR_SPECIAL},
 		{spc_illegal_op,},
 		{spc_illegal_op,},
@@ -359,7 +411,7 @@ static struct device_type_template sbc_template = {
 
 		{spc_illegal_op,},
 		{spc_illegal_op,},
-		{spc_mode_sense, NULL, PR_WE_FA|PR_EA_FA|PR_WE_FN|PR_EA_FN},
+		{sbc_mode_sense, NULL, PR_WE_FA|PR_EA_FA|PR_WE_FN|PR_EA_FN},
 		{spc_illegal_op,},
 		{spc_illegal_op,},
 		{spc_illegal_op,},
