@@ -852,7 +852,7 @@ static int spc_pr_read_reservation(int host_no, struct scsi_cmd *cmd)
 		else
 			res_key = reg->key;
 
-		put_unaligned_be32(res_key, &buf[8]);
+		put_unaligned_be64(res_key, &buf[8]);
 		buf[21] = ((reg->pr_scope << 4) & 0xf0) | (reg->pr_type & 0x0f);
 	} else
 		put_unaligned_be32(0, &buf[4]);
@@ -926,6 +926,19 @@ static struct registration *lookup_registration_by_nexus(struct scsi_lu *lu,
 	return NULL;
 }
 
+static int check_registration_key_exists(struct scsi_lu *lu, uint64_t key)
+{
+	struct registration *reg;
+
+	list_for_each_entry(reg, &lu->registration_list,
+				registration_siblings) {
+		if (reg->key == key)
+			return 0;
+	}
+
+	return 1;
+}
+
 static void __unregister(struct scsi_lu *lu, struct registration *reg)
 {
 	list_del(&reg->registration_siblings);
@@ -1192,9 +1205,8 @@ static int spc_pr_clear(int host_no, struct scsi_cmd *cmd)
 				 registration_siblings) {
 		/* we don't send myself */
 		if (sibling != reg)
-			ua_sense_add_other_it_nexus(sibling->nexus_id,
-						    cmd->dev,
-						    ASC_RESERVATIONS_PREEMPTED);
+			ua_sense_add_it_nexus(sibling->nexus_id,
+				cmd->dev, ASC_RESERVATIONS_PREEMPTED);
 		list_del(&sibling->registration_siblings);
 		free(sibling);
 	}
@@ -1211,10 +1223,11 @@ static int spc_pr_preempt(int host_no, struct scsi_cmd *cmd)
 	uint16_t asc = ASC_INVALID_FIELD_IN_CDB;
 	uint8_t key = ILLEGAL_REQUEST;
 	int ret, abort;
+	int res_released = 0, remove_all_reg = 0;
 	uint64_t res_key, sa_res_key;
 	uint8_t pr_scope, pr_type;
 	uint8_t *buf;
-	struct registration *reg, *sibling, *n;
+	struct registration *holder, *reg, *sibling, *n;
 
 	ret = check_pr_out_basic_parameter(cmd);
 	if (ret)
@@ -1237,56 +1250,59 @@ static int spc_pr_preempt(int host_no, struct scsi_cmd *cmd)
 	if (reg->key != res_key)
 		return SAM_STAT_RESERVATION_CONFLICT;
 
-remove_registration:
-	if (!cmd->dev->pr_holder) {
-		list_for_each_entry_safe(sibling, n, &cmd->dev->registration_list,
-					 registration_siblings) {
+	if (sa_res_key) {
+		ret = check_registration_key_exists(cmd->dev, sa_res_key);
+		if (ret)
+			return SAM_STAT_RESERVATION_CONFLICT;
+	}
 
-			if (sibling->key == sa_res_key) {
-				__unregister(cmd->dev, sibling);
-				break;
+	holder = cmd->dev->pr_holder;
+
+	if (holder) {
+		if (holder->pr_type == PR_TYPE_WRITE_EXCLUSIVE_ALLREG ||
+			holder->pr_type == PR_TYPE_EXCLUSIVE_ACCESS_ALLREG) {
+
+			if (!sa_res_key) {
+				if (pr_type != holder->pr_type ||
+					pr_scope != holder->pr_scope)
+					res_released = 1;
+				reg->pr_type = pr_type;
+				reg->pr_scope = pr_scope;
+				cmd->dev->pr_holder = reg;
+				remove_all_reg = 1;
+			}
+		} else {
+			if (holder->key == sa_res_key) {
+				if ((pr_type != holder->pr_type) ||
+					(pr_scope != holder->pr_scope))
+					res_released = 1;
+				reg->pr_type = pr_type;
+				reg->pr_scope = pr_scope;
+				cmd->dev->pr_holder = reg;
+			} else {
+				if (!sa_res_key)
+					goto sense;
 			}
 		}
-
-		return SAM_STAT_GOOD;
 	}
 
-	if (cmd->dev->pr_holder->pr_type == PR_TYPE_WRITE_EXCLUSIVE_ALLREG ||
-	    cmd->dev->pr_holder->pr_type == PR_TYPE_EXCLUSIVE_ACCESS_ALLREG) {
-
-		if (sa_res_key)
-			goto remove_registration;
-
-		list_for_each_entry_safe(sibling, n, &cmd->dev->registration_list,
-					 registration_siblings) {
-
-			if (sibling == reg)
-				continue;
+	list_for_each_entry_safe(sibling, n, &cmd->dev->registration_list,
+					registration_siblings) {
+		if (sibling == reg)
+			continue;
 
+		if (sibling->key == sa_res_key || remove_all_reg) {
+			ua_sense_add_it_nexus(sibling->nexus_id,
+				cmd->dev, ASC_RESERVATIONS_PREEMPTED);
 			__unregister(cmd->dev, sibling);
+		} else {
+			if (res_released)
+				ua_sense_add_it_nexus(sibling->nexus_id,
+				cmd->dev, ASC_RESERVATIONS_RELEASED);
 		}
-
-		cmd->dev->pr_holder = reg;
-		reg->pr_type = pr_type;
-		reg->pr_scope = pr_scope;
-
-		return SAM_STAT_GOOD;
 	}
 
-	if (cmd->dev->pr_holder->key == sa_res_key) {
-		cmd->dev->pr_holder = reg;
-		reg->pr_type = pr_type;
-		reg->pr_scope = pr_scope;
-
-		goto remove_registration;
-	}
-
-	if (sa_res_key)
-		goto remove_registration;
-
-	else
-		goto sense;
-
+	cmd->dev->prgeneration++;
 	return SAM_STAT_GOOD;
 sense:
 	scsi_set_in_resid_by_actual(cmd, 0);
@@ -1578,9 +1594,10 @@ enum {
 	Opt_scsi_id, Opt_scsi_sn,
 	Opt_vendor_id, Opt_product_id,
 	Opt_product_rev, Opt_sense_format,
-	Opt_removable, Opt_online,
+	Opt_removable, Opt_readonly, Opt_online,
 	Opt_mode_page,
 	Opt_path,
+	Opt_bsoflags,
 	Opt_err,
 };
 
@@ -1592,9 +1609,11 @@ static match_table_t tokens = {
 	{Opt_product_rev, "product_rev=%s"},
 	{Opt_sense_format, "sense_format=%s"},
 	{Opt_removable, "removable=%s"},
+	{Opt_readonly, "readonly=%s"},
 	{Opt_online, "online=%s"},
 	{Opt_mode_page, "mode_page=%s"},
 	{Opt_path, "path=%s"},
+	{Opt_bsoflags, "bsoflags=%s"},
 	{Opt_err, NULL},
 };
 
@@ -1627,6 +1646,7 @@ int lu_config(struct scsi_lu *lu, char *params, match_fn_t *fn)
 	while ((p = strsep(&params, ",")) != NULL) {
 		substring_t args[MAX_OPT_ARGS];
 		int token;
+
 		if (!*p)
 			continue;
 		token = match_token(p, tokens, args);
@@ -1661,6 +1681,10 @@ int lu_config(struct scsi_lu *lu, char *params, match_fn_t *fn)
 			match_strncpy(buf, &args[0], sizeof(buf));
 			attrs->removable = atoi(buf);
 			break;
+		case Opt_readonly:
+			match_strncpy(buf, &args[0], sizeof(buf));
+			attrs->readonly = atoi(buf);
+			break;
 		case Opt_online:
 			match_strncpy(buf, &args[0], sizeof(buf));
 			if (atoi(buf))
@@ -1719,6 +1743,7 @@ int spc_lu_init(struct scsi_lu *lu)
 	lu_vpd[pg]->vpd_update(lu, lu->attrs.scsi_id);
 
 	lu->attrs.removable = 0;
+	lu->attrs.readonly = 0;
 	lu->attrs.sense_format = 0;
 	lu->dev_type_template.lu_offline(lu);
 
