@@ -234,6 +234,33 @@ void ua_sense_add_other_it_nexus(uint64_t itn_id, struct scsi_lu *lu,
 	}
 }
 
+void ua_sense_add_it_nexus(uint64_t itn_id, struct scsi_lu *lu,
+				 uint16_t asc)
+{
+	struct it_nexus *itn;
+	struct it_nexus_lu_info *itn_lu;
+	int ret;
+
+	list_for_each_entry(itn, &lu->tgt->it_nexus_list, nexus_siblings) {
+
+		if (itn->itn_id == itn_id) {
+			list_for_each_entry(itn_lu, &itn->it_nexus_lu_info_list,
+					    lu_info_siblings) {
+
+				if (itn_lu->lu == lu) {
+					ret = ua_sense_add(itn_lu, asc);
+					if (ret)
+						eprintf("fail to add ua %"
+							PRIu64 " %" PRIu64
+							"\n", lu->lun, itn_id);
+					break;
+				}
+			}
+			break;
+		}
+	}
+}
+
 int it_nexus_create(int tid, uint64_t itn_id, int host_no, char *info)
 {
 	int i, ret;
@@ -418,12 +445,14 @@ __device_lookup(int tid, uint64_t lun, struct target **t)
 }
 
 enum {
-	Opt_path, Opt_bstype, Opt_err,
+	Opt_path, Opt_bstype, Opt_bsoflags, Opt_blocksize, Opt_err,
 };
 
 static match_table_t device_tokens = {
 	{Opt_path, "path=%s"},
 	{Opt_bstype, "bstype=%s"},
+	{Opt_bsoflags, "bsoflags=%s"},
+	{Opt_blocksize, "blocksize=%s"},
 	{Opt_err, NULL},
 };
 
@@ -433,7 +462,8 @@ int tgt_device_create(int tid, int dev_type, uint64_t lun, char *params,
 		      int backing)
 {
 	char *p, *path = NULL, *bstype = NULL;
-	int ret = 0;
+	char *bsoflags = NULL, *blocksize = NULL;
+	int ret = 0, lu_bsoflags = 0;
 	struct target *target;
 	struct scsi_lu *lu, *pos;
 	struct device_type_template *t;
@@ -456,6 +486,12 @@ int tgt_device_create(int tid, int dev_type, uint64_t lun, char *params,
 		case Opt_bstype:
 			bstype = match_strdup(&args[0]);
 			break;
+		case Opt_bsoflags:
+			bsoflags = match_strdup(&args[0]);
+			break;
+		case Opt_blocksize:
+			blocksize = match_strdup(&args[0]);
+			break;
 		default:
 			break;
 		}
@@ -487,6 +523,27 @@ int tgt_device_create(int tid, int dev_type, uint64_t lun, char *params,
 	} else
 		bst = get_backingstore_template("null");
 
+	if ((!strncmp(bst->bs_name, "bsg", 3) ||
+	     !strncmp(bst->bs_name, "sg", 2)) &&
+	    dev_type != TYPE_PT) {
+		ret = TGTADM_INVALID_REQUEST;
+		goto out;
+	}
+
+	if (bsoflags) {
+		lu_bsoflags = str_to_open_flags(bsoflags);
+		if (lu_bsoflags == -1) {
+			ret = TGTADM_INVALID_REQUEST;
+			goto out;
+		}
+	}
+
+	if (lu_bsoflags && !(bst->bs_oflags_supported & lu_bsoflags)) {
+		eprintf("bsoflags not supported\n");
+		ret = TGTADM_INVALID_REQUEST;
+		goto out;
+	}
+
 	t = device_type_lookup(dev_type);
 	if (!t) {
 		eprintf("Unknown device type %d\n", dev_type);
@@ -504,6 +561,8 @@ int tgt_device_create(int tid, int dev_type, uint64_t lun, char *params,
 	lu->bst = bst;
 	lu->tgt = target;
 	lu->lun = lun;
+	lu->bsoflags = lu_bsoflags;
+
 	tgt_cmd_queue_init(&lu->cmd_queue);
 	INIT_LIST_HEAD(&lu->registration_list);
 	lu->prgeneration = 0;
@@ -512,7 +571,26 @@ int tgt_device_create(int tid, int dev_type, uint64_t lun, char *params,
 	lu->cmd_perform = &target_cmd_perform;
 	lu->cmd_done = &__cmd_done;
 
- 	if (lu->dev_type_template.lu_init) {
+	lu->blk_shift = 0;
+	if (blocksize) {
+		unsigned int bsize;
+		unsigned int bshift;
+
+		dprintf("blocksize=%s\n", blocksize);
+		bsize = strtoul(blocksize, NULL, 0);
+
+		bshift = get_blk_shift(bsize);
+		if (bshift > 0)
+			lu->blk_shift = bshift;
+		else {
+			if (bsize > 0)
+				eprintf("%u is invalid block size\n", bsize);
+			else
+				eprintf("%s is invalid block size\n", blocksize);
+		}
+	}
+
+	if (lu->dev_type_template.lu_init) {
 		ret = lu->dev_type_template.lu_init(lu);
 		if (ret)
 			goto fail_lu_init;
@@ -571,8 +649,12 @@ int tgt_device_create(int tid, int dev_type, uint64_t lun, char *params,
 out:
 	if (bstype)
 		free(bstype);
+	if (blocksize)
+		free(blocksize);
 	if (path)
 		free(path);
+	if (bsoflags)
+		free(bsoflags);
 	return ret;
 fail_bs_init:
 	if (lu->bst->bs_exit)
@@ -1637,7 +1719,8 @@ static struct {
 	{TYPE_ENCLOSURE, "enclosure"},
 	{TYPE_RBC, "rbc"},
 	{TYPE_OSD, "osd"},
-	{TYPE_NO_LUN, "No LUN"}
+	{TYPE_NO_LUN, "No LUN"},
+	{TYPE_PT, "passthrough"}
 };
 
 static char *print_type(int type)
@@ -1654,10 +1737,10 @@ static char *print_type(int type)
 	return name;
 }
 
-
 int tgt_target_show_all(char *buf, int rest)
 {
 	int total = 0, max = rest;
+	char strflags[128];
 	struct target *target;
 	struct scsi_lu *lu;
 	struct acl_entry *acl;
@@ -1690,22 +1773,27 @@ int tgt_target_show_all(char *buf, int rest)
   				 _TAB3 "Type: %s\n"
 				 _TAB3 "SCSI ID: %s\n"
 				 _TAB3 "SCSI SN: %s\n"
-				 _TAB3 "Size: %s\n"
+				 _TAB3 "Size: %s, Block size: %d\n"
 				 _TAB3 "Online: %s\n"
 				 _TAB3 "Removable media: %s\n"
+				 _TAB3 "Readonly: %s\n"
 				 _TAB3 "Backing store type: %s\n"
-				 _TAB3 "Backing store path: %s\n",
+				 _TAB3 "Backing store path: %s\n"
+				 _TAB3 "Backing store flags: %s\n",
 				 lu->lun,
   				 print_type(lu->attrs.device_type),
 				 lu->attrs.scsi_id,
 				 lu->attrs.scsi_sn,
 				 print_disksize(lu->size),
+				 1U << lu->blk_shift,
 				 lu->attrs.online ? "Yes" : "No",
 				 lu->attrs.removable ? "Yes" : "No",
+				 lu->attrs.readonly ? "Yes" : "No",
 				 lu->bst ?
 					(lu->bst->bs_name ? : "Unknown") :
 					"None",
-				 lu->path ? : "None");
+				 lu->path ? : "None",
+				 open_flags_to_str(strflags, lu->bsoflags));
 
 		if (!strcmp(tgt_drivers[target->lid]->name, "iscsi")) {
 			int i, aid;
