@@ -41,19 +41,8 @@
 #include "list.h"
 #include "tgtadm.h"
 
-#undef eprintf
-#define eprintf(fmt, args...)						\
-do {									\
-	fprintf(stderr, "%s: " fmt, program_name, ##args);		\
-} while (0)
-
-#undef dprintf
-#define dprintf(fmt, args...)						\
-do {									\
-	if (debug)							\
-		fprintf(stderr, "%s %d: " fmt,				\
-			__FUNCTION__, __LINE__, ##args);		\
-} while (0)
+#define NO_LOGGING
+#include "log.h"
 
 #define BUFSIZE 4096
 
@@ -101,6 +90,7 @@ static const char * tgtadm_strerror(int err)
 struct option const long_options[] = {
 	{"debug", no_argument, NULL, 'd'},
 	{"help", no_argument, NULL, 'h'},
+	{"version", no_argument, NULL, 'V'},
 	{"lld", required_argument, NULL, 'L'},
 	{"op", required_argument, NULL, 'o'},
 	{"mode", required_argument, NULL, 'm'},
@@ -112,6 +102,7 @@ struct option const long_options[] = {
 	{"value", required_argument, NULL, 'v'},
 	{"backing-store", required_argument, NULL, 'b'},
 	{"bstype", required_argument, NULL, 'E'},
+	{"bsoflags", required_argument, NULL, 'f'},
 	{"targetname", required_argument, NULL, 'T'},
 	{"initiator-address", required_argument, NULL, 'I'},
 	{"user", required_argument, NULL, 'u'},
@@ -126,7 +117,7 @@ struct option const long_options[] = {
 	{NULL, 0, NULL, 0},
 };
 
-static char *short_options = "dhL:o:m:t:s:c:l:n:v:b:E:T:I:u:p:H:P:B:Y:O:C:";
+static char *short_options = "dhVL:o:m:t:s:c:l:n:v:b:E:f:T:I:u:p:H:P:B:Y:O:C:";
 
 static void usage(int status)
 {
@@ -137,45 +128,48 @@ static void usage(int status)
 		printf("\
 Linux SCSI Target Framework Administration Utility, version %s\n\
 \n\
-  --lld [driver] --mode target --op new --tid=[id] --targetname [name]\n\
-                        add a new target with [id] and [name]. [id] must not be zero.\n\
-  --lld [driver] --mode target --op delete --tid=[id]\n\
-                        delete the specific target with [id]. The target must\n\
+  --lld <driver> --mode target --op new --tid <id> --targetname <name>\n\
+                        add a new target with <id> and <name>. <id> must not be zero.\n\
+  --lld <driver> --mode target --op delete --tid <id>\n\
+                        delete the specific target with <id>. The target must\n\
                         have no activity.\n\
-  --lld [driver] --mode target --op show\n\
+  --lld <driver> --mode target --op show\n\
                         show all the targets.\n\
-  --lld [driver] --mode target --op show --tid=[id]\n\
+  --lld <driver> --mode target --op show --tid <id>\n\
                         show the specific target's parameters.\n\
-  --lld [driver] --mode target --op update --tid=[id] --name=[param] --value=[value]\n\
+  --lld <driver> --mode target --op update --tid <id> --name <param> --value <value>\n\
                         change the target parameters of the specific\n\
-                        target with [id].\n\
-  --lld [driver] --mode target --op bind --tid=[id] --initiator-address=[src]\n\
+                        target with <id>.\n\
+  --lld <driver> --mode target --op bind --tid <id> --initiator-address <src>\n\
                         enable the target to accept the specific initiators.\n\
-  --lld [driver] --mode target --op unbind --tid=[id] --initiator-address=[src]\n\
+  --lld <driver> --mode target --op unbind --tid <id> --initiator-address <src>\n\
                         disable the specific permitted initiators.\n\
-  --lld [driver] --mode logicalunit --op new --tid=[id] --lun=[lun] --backing-store=[path] --bstype=[type]\n\
-                        add a new logical unit with [lun] to the specific\n\
-                        target with [id]. The logical unit is offered\n\
-                        to the initiators. [path] must be block device files\n\
+  --lld <driver> --mode logicalunit --op new --tid <id> --lun <lun> \\\n\
+                        --backing-store <path> --bstype <type> --bsoflags <options>\n\
+                        add a new logical unit with <lun> to the specific\n\
+                        target with <id>. The logical unit is offered\n\
+                        to the initiators. <path> must be block device files\n\
                         (including LVM and RAID devices) or regular files.\n\
                         bstype option is optional.\n\
-  --lld [driver] --mode logicalunit --op delete --tid=[id] --lun=[lun]\n\
-                        delete the specific logical unit with [lun] that\n\
-                        the target with [id] has.\n\
-  --lld [driver] --mode account --op new --user=[name] --password=[pass]\n\
-                        add a new account with [name] and [pass].\n\
-  --lld [driver] --mode account --op delete --user=[name]\n\
-                        delete the specific account having [name].\n\
-  --lld [driver] --mode account --op bind --tid=[id] --user=[name] [--outgoing]\n\
-                        add the specific account having [name] to\n\
-                        the specific target with [id].\n\
-                        [user] could be [IncomingUser] or [OutgoingUser].\n\
+                        bsoflags supported options are sync and direct\n\
+                        (sync:direct for both).\n\
+  --lld <driver> --mode logicalunit --op delete --tid <id> --lun <lun>\n\
+                        delete the specific logical unit with <lun> that\n\
+                        the target with <id> has.\n\
+  --lld <driver> --mode account --op new --user <name> --password <pass>\n\
+                        add a new account with <name> and <pass>.\n\
+  --lld <driver> --mode account --op delete --user <name>\n\
+                        delete the specific account having <name>.\n\
+  --lld <driver> --mode account --op bind --tid <id> --user <name> [--outgoing]\n\
+                        add the specific account having <name> to\n\
+                        the specific target with <id>.\n\
+                        <user> could be <IncomingUser> or <OutgoingUser>.\n\
                         If you use --outgoing option, the account will\n\
                         be added as an outgoing account.\n\
-  --lld [driver] --mode account --op unbind --tid=[id] --user=[name]\n\
-                        delete the specific account having [name] from specific\n\
+  --lld <driver> --mode account --op unbind --tid <id> --user <name>\n\
+                        delete the specific account having <name> from specific\n\
                         target.\n\
-  --control-port NNNN   use control port NNNN\n\
+  --control-port <port> use control port <port>\n\
   --help                display this help and exit\n\
 \n\
 Report bugs to <stgt@vger.kernel.org>.\n", TGT_VERSION);
@@ -183,6 +177,12 @@ Report bugs to <stgt@vger.kernel.org>.\n", TGT_VERSION);
 	exit(status == 0 ? 0 : EINVAL);
 }
 
+static void version(void)
+{
+	printf("%s\n", TGT_VERSION);
+	exit(0);
+}
+
 /* default port to use for the mgmt channel */
 static short int control_port = 0;
 
@@ -379,6 +379,8 @@ static int str_to_mode(char *str)
 		return MODE_CONNECTION;
 	else if (!strcmp("account", str))
 		return MODE_ACCOUNT;
+	else if (!strcmp("stats", str))
+		return MODE_STATS;
 	else {
 		eprintf("unknown mode: %s\n", str);
 		exit(1);
@@ -431,6 +433,7 @@ int main(int argc, char **argv)
 	uint64_t sid, lun;
 	char *name, *value, *path, *targetname, *params, *address, *targetOps;
 	char *bstype;
+	char *bsoflags;
 	char *user, *password;
 	char *buf;
 	size_t bufsz = BUFSIZE + sizeof(struct tgtadm_req);
@@ -445,7 +448,7 @@ int main(int argc, char **argv)
 	ac_dir = ACCOUNT_TYPE_INCOMING;
 	rest = BUFSIZE;
 	name = value = path = targetname = address = targetOps = bstype = NULL;
-	user = password = NULL;
+	bsoflags = user = password = NULL;
 
 	buf = valloc(bufsz);
 	if (!buf) {
@@ -516,6 +519,9 @@ int main(int argc, char **argv)
 		case 'H':
 			hostno = strtol(optarg, NULL, 10);
 			break;
+		case 'f':
+			bsoflags = optarg;
+			break;
 		case 'E':
 			bstype = optarg;
 			break;
@@ -528,6 +534,9 @@ int main(int argc, char **argv)
 		case 'C':
 			control_port = strtol(optarg, NULL, 10);
 			break;
+		case 'V':
+			version();
+			break;
 		case 'd':
 			debug = 1;
 			break;
@@ -644,7 +653,7 @@ int main(int argc, char **argv)
 	if (mode == MODE_ACCOUNT) {
 		switch (op) {
 		case OP_NEW:
-			rc = verify_mode_params(argc, argv, "LmoupC");
+			rc = verify_mode_params(argc, argv, "LmoupfC");
 			if (rc) {
 				eprintf("logicalunit mode: option '-%c' is not "
 					  "allowed/supported\n", rc);
@@ -686,7 +695,7 @@ int main(int argc, char **argv)
 				tid = GLOBAL_TID;
 			break;
 		case OP_UNBIND:
-			rc = verify_mode_params(argc, argv, "LmouC");
+			rc = verify_mode_params(argc, argv, "LmotuC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
 					  "allowed/supported\n", rc);
@@ -717,7 +726,7 @@ int main(int argc, char **argv)
 		}
 		switch (op) {
 		case OP_NEW:
-			rc = verify_mode_params(argc, argv, "LmotlbEYC");
+			rc = verify_mode_params(argc, argv, "LmoftlbEYC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
 					  "allowed/supported\n", rc);
@@ -738,7 +747,7 @@ int main(int argc, char **argv)
 			}
 			break;
 		case OP_UPDATE:
-			rc = verify_mode_params(argc, argv, "LmotlPC");
+			rc = verify_mode_params(argc, argv, "LmoftlPC");
 			if (rc) {
 				eprintf("option '-%c' not supported in "
 					"logicalunit mode\n", rc);
@@ -777,6 +786,9 @@ int main(int argc, char **argv)
 	else if (bstype)
 		shprintf(total, params, rest, "%sbstype=%s",
 			 rest == BUFSIZE ? "" : ",", bstype);
+	if (bsoflags)
+		shprintf(total, params, rest, "%sbsoflags=%s",
+			 rest == BUFSIZE ? "" : ",", bsoflags);
 	if (targetname)
 		shprintf(total, params, rest, "%stargetname=%s",
 			 rest == BUFSIZE ? "" : ",", targetname);
