@@ -60,9 +60,10 @@ struct lu_phy_attr {
 
 	uint16_t version_desc[VERSION_DESCRIPTOR_LEN];
 
-	char device_type;	/* Peripheral device type */
+	unsigned char device_type; /* Peripheral device type */
 	char qualifier;		/* Peripheral Qualifier */
 	char removable;		/* Removable media */
+	char readonly;          /* Read-Only media */
 	char online;		/* Logical Unit online */
 	char sense_format;	/* Descrptor format sense data supported */
 
@@ -124,6 +125,7 @@ struct backingstore_template {
 	void (*bs_exit)(struct scsi_lu *dev);
 	int (*bs_cmd_submit)(struct scsi_cmd *cmd);
 	int (*bs_cmd_done)(struct scsi_cmd *cmd);
+	int bs_oflags_supported;
 
 	struct list_head backingstore_siblings;
 };
@@ -151,6 +153,8 @@ struct scsi_lu {
 	uint64_t size;
 	uint64_t lun;
 	char *path;
+	int bsoflags;
+	unsigned int blk_shift;
 
 	/* the list of devices belonging to a target */
 	struct list_head device_siblings;
@@ -279,6 +283,8 @@ extern int ua_sense_del(struct scsi_cmd *cmd, int del);
 extern void ua_sense_clear(struct it_nexus_lu_info *itn_lu, uint16_t asc);
 extern void ua_sense_add_other_it_nexus(uint64_t itn_id, struct scsi_lu *lu,
 					uint16_t asc);
+extern void ua_sense_add_it_nexus(uint64_t itn_id, struct scsi_lu *lu,
+					uint16_t asc);
 
 extern uint64_t scsi_get_devid(int lid, uint8_t *pdu);
 extern int scsi_cmd_perform(int host_no, struct scsi_cmd *cmd);
@@ -332,21 +338,7 @@ struct event_data {
 	struct list_head e_list;
 };
 
-extern int sig_fd;
-
-struct bs_finish {
-	struct list_head finished_list;
-	pthread_mutex_t finished_lock;
-	pthread_t thread;
-};
-
-void bs_sig_request_done(int fd, int events, void *data);
-
-int do_tgt_event_add(int efd, struct list_head *list, int fd, int events,
-		     event_handler_t handler, void *data);
-void do_tgt_event_del(int efd, struct list_head *list, int fd);
-int do_tgt_event_modify(int efd, struct list_head *list, int fd, int events);
-
-int run_ext_program(const char *cmd,
-		    void (*callback)(void *data, int result), void *data);
+int call_program(const char *cmd,
+		    void (*callback)(void *data, int result), void *data,
+		    char *output, int op_len, int flags);
 #endif
